@@ -1,23 +1,21 @@
 #![cfg_attr(docsrs, feature(doc_cfg))]
 
-//! A fast, asynchronous terminal paging library for Rust. `minus` provides high
-//! level functions to easily embed a pager for any terminal application.
+//! `minus` is a terminal paging library for Rust. It provides high
+//! level functions to easily embed a pager for any terminal application without
+//! requiring you to think about the nitty-gritty of building a pager.
 //!
-//! `minus` can be used in asynchronous mode or in a blocking fashion
+//! `minus` can be used in asynchronous mode or in a blocking fashion.
 //!
-//! * In asynchronous mode, the pager's data as well as it's
-//! configuration can be **updated** at any time.`minus` supports both
-//! [`tokio`] as well as [`async-std`] runtimes. The support
-//! for these runtimes are gated on individual features.
+//! In asynchronous mode, the pager's data as well as it's
+//! configuration can be **updated** at any time. `minus` supports both
+//! [`tokio`] as well as [`async-std`] runtimes. These runtimes are gated
+//! on specific [features](./index.html#features).
 //!
-//! * In blocking mode, the pager stops any other code from being executed. This
+//! In blocking mode, the pager stops any other task from being executed. This
 //! is good if you want to show some static information but it does not allow
 //! you to change the configuration of the pager at runtime.
 //!
-//! * When using `minus`, you select what features you need and **nothing else**.
-//!
 //! # Features
-//!
 //! * `async_std_lib`: Use this if you use [`async_std`] runtime in your
 //! application
 //! * `tokio_lib`:Use this if you are using [`tokio`] runtime for your application
@@ -25,6 +23,10 @@
 //! output
 //! * `search`: If you want searching capablities inside the feature
 //!
+//! **Note:** You must select a either a runtime feature or `static_output`. If you do not
+//! select any features when adding `minus` to `Cargo.toml`, it will not have any functions
+//! to start a pager.
+//!
 //! # Examples
 //! Print numbers 1 through 100 with 100ms delay in asynchronous mode
 //!
@@ -37,24 +39,24 @@
 //!
 //! #[tokio::main]
 //! async fn main() -> Result<(), Box<dyn std::error::Error>> {
-//!     let mut pager = Pager::new().unwrap();
+//!     let mut pager = Pager::new()?;
 //!     pager.set_prompt("An asynchronous example");
-//!     let pager = pager.finish();
+//!     let pager_mutex = pager.finish();
 //!
 //!     let updater = async {
 //!         for i in 1..=100u8 {
-//!             let mut guard = pager.lock().await;
+//!             let mut guard = pager_mutex.lock().await;
 //!             writeln!(guard, "{}", i)?;
 //!             // Remember to drop the guard before any await or blocking operation
 //!             drop(guard);
 //!             sleep(Duration::from_millis(100)).await;
 //!         }
-//!         let mut guard = pager.lock().await;
+//!         let mut guard = pager_mutex.lock().await;
 //!         guard.end_data_stream();
 //!         Result::<_, std::fmt::Error>::Ok(())
 //!     };
 //!
-//!     let (res1, res2) = join!(tokio_updating(pager.clone()), updater);
+//!     let (res1, res2) = join!(tokio_updating(pager_mutex.clone()), updater);
 //!     res1?;
 //!     res2?;
 //!     Ok(())
@@ -118,8 +120,8 @@ pub use rt_wrappers::*;
 pub use search::SearchMode;
 #[cfg(feature = "static_output")]
 pub use static_pager::page_all;
+use std::string::ToString;
 use std::{fmt, io::stdout};
-use std::{iter::Flatten, string::ToString, vec::IntoIter};
 pub use utils::LineNumbers;
 
 #[cfg(any(feature = "tokio_lib", feature = "async_std_lib"))]
@@ -127,9 +129,9 @@ pub use utils::LineNumbers;
     docsrs,
     doc(cfg(any(feature = "tokio_lib", feature = "async_std_lib")))
 )]
-/// A convenience type for `std::sync::Arc<async_mutex::Mutex<Pager>>`
+/// A convenient type for `std::sync::Arc<async_mutex::Mutex<Pager>>`
 pub type PagerMutex = std::sync::Arc<Mutex<Pager>>;
-/// A convenience type for `Vec<Box<dyn FnMut() + Send + Sync + 'static>>`
+/// A convenient type for `Vec<Box<dyn FnMut() + Send + Sync + 'static>>`
 pub type ExitCallbacks = Vec<Box<dyn FnMut() + Send + Sync + 'static>>;
 
 // The Wrapping Model
@@ -137,16 +139,16 @@ pub type ExitCallbacks = Vec<Box<dyn FnMut() + Send + Sync + 'static>>;
 // minus heavily uses the wrapping model. This is key to understand how minus
 // internally.
 //
-// When a text is given to minus in for displaying, it internally takes each
-// logical line of it and breaks it into a `Vec<String>`. To hold multiple of
+// When a text is given to minus for displaying, it internally takes each
+// logical line of it and breaks it into a `Vec<String>`. To hold all of
 // those lines, it stores them inside another Vec container. This makes it the
 // `Vec<Vec<String>>` struct.
 //
 // Each element in the 1st `Vec` is a logical line. While each String in the 2nd
 // `Vec` is a line wrapped to the available terminal width.
 //
-// In case of prompt text and message, which are allowed to take only one row in
-// the terminal, and hence, must contain only 1 line are contained in a
+// In case of prompt text and message, which are allowed to occupy only a single
+//  row in the terminal, and hence, must contain only 1 line are contained in a
 // `Vec<String>`
 //
 // If the terminal is resized, we update the rows and columns and rewrap the
@@ -156,15 +158,14 @@ pub type ExitCallbacks = Vec<Box<dyn FnMut() + Send + Sync + 'static>>;
 ///
 /// This is used by all initializing functions
 pub struct Pager {
-    // The output that is displayed wrapped to the available terminal width
-    wrap_lines: Vec<Vec<String>>,
+    // The text the pager has been told to be displayed
+    lines: String,
+    // The output, flattened and formatted into the lines that should be displayed
+    formatted_lines: Vec<String>,
     // Configuration for line numbers. See [`LineNumbers`]
     pub(crate) line_numbers: LineNumbers,
     // The prompt displayed at the bottom wrapped to available terminal width
     prompt: Vec<String>,
-    // Text which may have come through `push_str` (or `writeln`) that isn't
-    // flushed to wrap_lines, since it isn't terminated yet with a \n
-    lines: String,
     // The input classifier to be called when a input is found
     input_classifier: Box<dyn input::InputClassifier + Sync + Send>,
     // Functions to run when the pager quits
@@ -221,8 +222,8 @@ impl Pager {
         } else if stdout().is_tty() {
             // If a proper terminal is present, get size and set it
             let size = terminal::size()?;
-            cols = size.0;
-            rows = size.1;
+            cols = size.0 as usize;
+            rows = size.1 as usize;
         } else {
             // For other cases beyond control
             cols = 1;
@@ -230,16 +231,16 @@ impl Pager {
         };
 
         Ok(Pager {
-            wrap_lines: Vec::new(),
+            lines: String::new(),
+            formatted_lines: Vec::new(),
             line_numbers: LineNumbers::Disabled,
             upper_mark: 0,
-            prompt: wrap_str("minus", cols.into()),
+            prompt: wrap_str("minus", cols),
             exit_strategy: ExitStrategy::ProcessQuit,
-            input_classifier: Box::new(input::DefaultInputHandler {}),
+            input_classifier: Box::new(input::DefaultInputClassifier {}),
             exit_callbacks: Vec::new(),
             run_no_overflow: false,
             message: (None, false),
-            lines: String::new(),
             end_stream: false,
             #[cfg(feature = "search")]
             search_term: None,
@@ -248,8 +249,8 @@ impl Pager {
             #[cfg(feature = "search")]
             search_idx: Vec::new(),
             // Just to be safe in tests, keep at 1x1 size
-            cols: cols as usize,
-            rows: rows as usize,
+            cols,
+            rows,
         })
     }
 
@@ -264,9 +265,8 @@ impl Pager {
     /// pager.set_text("This is a line");
     /// ```
     pub fn set_text(&mut self, text: impl Into<String>) {
-        let text: String = text.into();
-        // self.lines = WrappedLines::from(Line::from_str(&text.into(), self.cols));
-        self.wrap_lines = text.lines().map(|l| wrap_str(l, self.cols)).collect();
+        self.lines = text.into();
+        self.format_lines();
     }
 
     /// Set line number to this setting
@@ -280,6 +280,7 @@ impl Pager {
     /// ```
     pub fn set_line_numbers(&mut self, l: LineNumbers) {
         self.line_numbers = l;
+        self.format_lines();
     }
 
     /// Display a temporary message at the prompt area
@@ -363,14 +364,6 @@ impl Pager {
         self.exit_strategy = strategy;
     }
 
-    /// Returns the appropriate text for displaying.
-    ///
-    /// Nrmally it will return `self.lines`
-    /// In case of a search, `self.search_lines` is returned
-    pub(crate) fn get_lines(&self) -> Vec<Vec<String>> {
-        self.wrap_lines.clone()
-    }
-
     /// Set whether to display pager if there's less data than
     /// available screen height
     ///
@@ -396,31 +389,67 @@ impl Pager {
     /// pager.push_str("This is some text");
     /// ```
     pub fn push_str(&mut self, string: impl Into<String>) {
-        let string = string.into();
-        if string.ends_with('\n') {
-            self.lines.push_str(&string);
-            self.wrap_lines.append(
-                &mut self
-                    .lines
-                    .lines()
-                    .map(|l| wrap_str(l, self.cols))
-                    .collect::<Vec<Vec<String>>>(),
-            );
-            self.lines.clear();
-        } else if string.contains('\n') {
-            let mut lines = string.lines().collect::<Vec<&str>>();
-            let line_count = lines.len();
-            let push_lines = &mut lines[0..line_count - 1];
-            self.wrap_lines.append(
-                &mut push_lines
-                    .iter()
-                    .map(|l| wrap_str(l, self.cols))
-                    .collect::<Vec<Vec<String>>>(),
-            );
-            self.lines.push_str(lines[line_count - 1]);
+        let text = string.into();
+
+        // if the text we have saved currently ends with a newline,
+        // we want the formatted_text vector to append the line instead of
+        // trying to add it to the last item
+        let newline = self.lines.ends_with('\n');
+
+        // find the number of trailing whitespace characters currently on self.lines
+        let ending_whitespace = self
+            .lines
+            .chars()
+            .rev()
+            .take_while(|c| c.is_whitespace() && *c != '\n')
+            .collect::<String>();
+
+        // push the text to lines
+        self.lines.push_str(&text);
+
+        // And get how many lines of text will be shown (not how many rows, how many wrapped
+        // lines), and get its string length
+        let len_line_number = self.lines.lines().count().to_string().len();
+
+        // if we want a newline, just format the new text and append it.
+        // if we don't, format the text with the last line currently formatted
+        // since it will be appended to that
+        //
+        // also get the line number to start at when formatting
+        let (to_format, to_skip) = if newline {
+            (text, self.lines.lines().count())
         } else {
-            self.lines.push_str(&string);
-        }
+            // add the trailing whitespace in here, since it isn't preserved when wrapping the
+            // lines, and thus won't appear on the last element in self.formatted_lines
+            let to_fmt = format!(
+                "{}{}{}",
+                self.formatted_lines.pop().unwrap_or_default(),
+                ending_whitespace,
+                text
+            );
+
+            (to_fmt, self.lines.lines().count().saturating_sub(1))
+        };
+
+        // format the lines we want to format
+        let mut to_append = to_format
+            .lines()
+            .enumerate()
+            .flat_map(|(idx, line)| {
+                self.formatted_line(
+                    line,
+                    matches!(
+                        self.line_numbers,
+                        LineNumbers::AlwaysOn | LineNumbers::Enabled
+                    ),
+                    len_line_number,
+                    idx + to_skip,
+                )
+            })
+            .collect::<Vec<String>>();
+
+        // append the new vector to the formatted lines
+        self.formatted_lines.append(&mut to_append);
     }
 
     /// Hints the running pager that no more data is coming
@@ -437,28 +466,103 @@ impl Pager {
         self.end_stream = true;
     }
 
-    /// Readjust the text to new terminal size
-    pub(crate) fn readjust_wraps(&mut self) {
-        rewrap_lines(&mut self.wrap_lines, self.cols);
+    pub(crate) fn formatted_line(
+        &self,
+        line: &str,
+        line_numbers: bool,
+        len_line_number: usize,
+        idx: usize,
+    ) -> Vec<String> {
+        if line_numbers {
+            #[cfg_attr(not(feature = "search"), allow(unused_mut))]
+            wrap_str(line, self.cols.saturating_sub(len_line_number + 3))
+                .into_iter()
+                .map(|mut row| {
+                    #[cfg(feature = "search")]
+                    if let Some(st) = self.search_term.as_ref() {
+                        // highlight the lines with matching search terms
+                        row = search::highlight_line_matches(&row, st);
+                    }
+
+                    if cfg!(not(test)) {
+                        format!(
+                            " {bold}{number: >len$}.{reset} {row}",
+                            bold = crossterm::style::Attribute::Bold,
+                            number = idx + 1,
+                            len = len_line_number,
+                            reset = crossterm::style::Attribute::Reset,
+                            row = row
+                        )
+                    } else {
+                        format!(
+                            " {number: >len$}. {row}",
+                            number = idx + 1,
+                            len = len_line_number,
+                            row = row
+                        )
+                    }
+                })
+                .collect::<Vec<String>>()
+        } else {
+            #[cfg(feature = "search")]
+            if let Some(st) = self.search_term.as_ref() {
+                return wrap_str(&search::highlight_line_matches(line, st), self.cols);
+            }
+
+            wrap_str(line, self.cols)
+        }
+    }
+
+    pub(crate) fn format_lines(&mut self) {
+        let line_count = self.lines.lines().count();
+
+        let len_line_number = line_count.to_string().len();
+
+        self.formatted_lines = self
+            .lines
+            .lines()
+            .enumerate()
+            .flat_map(|(idx, line)| {
+                self.formatted_line(
+                    line,
+                    matches!(
+                        self.line_numbers,
+                        LineNumbers::AlwaysOn | LineNumbers::Enabled
+                    ),
+                    len_line_number,
+                    idx,
+                )
+            })
+            .collect::<Vec<String>>();
+
         if self.message.0.is_some() {
             rewrap(&mut self.message.0.as_mut().unwrap(), self.cols);
         }
         rewrap(&mut self.prompt, self.cols);
+
+        #[cfg(feature = "search")]
+        search::set_match_indices(self);
     }
 
-    /// Returns all the text by flattening them into a single vector of strings
-    pub(crate) fn get_flattened_lines(&self) -> Flatten<IntoIter<Vec<String>>> {
-        self.get_lines().into_iter().flatten()
+    /// Returns all the text within the bounds, after flattening
+    pub(crate) fn get_flattened_lines_with_bounds(&self, start: usize, end: usize) -> &[String] {
+        if start >= self.num_lines() || start > end {
+            &[]
+        } else if end >= self.num_lines() {
+            &self.formatted_lines[start..]
+        } else {
+            &self.formatted_lines[start..end]
+        }
     }
 
     /// Returns the number of lines the [`Pager`] currently holds
     pub(crate) fn num_lines(&self) -> usize {
-        self.get_flattened_lines().count()
+        self.formatted_lines.len()
     }
 
     /// Set custom input handler function
     ///
-    /// See example in [`InputHandler`](input::InputHandler) on using this
+    /// See example in [`InputHandler`](input::InputClassifier) on using this
     /// function
     pub fn set_input_handler(&mut self, handler: Box<dyn input::InputClassifier + Send + Sync>) {
         self.input_classifier = handler;
@@ -512,13 +616,6 @@ pub enum ExitStrategy {
     PagerQuit,
 }
 
-/// Rewrap already wrapped vector of lines based on the number of columns
-pub(crate) fn rewrap_lines(lines: &mut Vec<Vec<String>>, cols: usize) {
-    for line in lines {
-        rewrap(line, cols);
-    }
-}
-
 /// Rewrap a single line based on the number of columns
 pub(crate) fn rewrap(line: &mut Vec<String>, cols: usize) {
     *line = textwrap::wrap(&line.join(" "), cols)
