@@ -5,15 +5,18 @@
 use crate::error::AlternateScreenPagingError;
 use crate::init;
 use crate::PagerMutex;
-use std::sync::Arc;
 
 #[cfg(feature = "async_std_lib")]
-pub mod async_std_wrapper;
+mod async_std_wrapper;
+#[cfg(feature = "async_std_lib")]
+pub use async_std_wrapper::async_std_updating;
 
 #[cfg(feature = "tokio_lib")]
-pub mod tokio_wrapper;
+mod tokio_wrapper;
+#[cfg(feature = "tokio_lib")]
+pub use tokio_wrapper::tokio_updating;
 
 /// Private function that contains the implemenation for the async display.
-async fn run(pager: Arc<PagerMutex>) -> Result<(), AlternateScreenPagingError> {
+async fn run(pager: PagerMutex) -> Result<(), AlternateScreenPagingError> {
     init::dynamic_paging(&pager).await
 }
